@@ -1,13 +1,13 @@
 ;;; ellama.el --- Tool for interacting with LLMs -*- lexical-binding: t -*-
 
-;; Copyright (C) 2023-2025  Free Software Foundation, Inc.
+;; Copyright (C) 2023-2026  Free Software Foundation, Inc.
 
 ;; Author: Sergey Kostyaev <sskostyaev@gmail.com>
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
 ;; Package-Requires: ((emacs "28.1") (llm "0.24.0") (plz "0.8") (transient "0.7") (compat "29.1"))
-;; Package-Version: 20250918.1606
-;; Package-Revision: 2857f85b8f10
+;; Package-Version: 20260119.2229
+;; Package-Revision: c491bf4c5b64
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -41,6 +41,7 @@
 (require 'llm-provider-utils)
 (require 'compat)
 (eval-when-compile (require 'rx))
+(require 'ellama-tools)
 
 (defgroup ellama nil
   "Tool for interacting with LLMs."
@@ -226,6 +227,15 @@ PROMPT is a prompt string."
 	  (const :tag "By generating name with reasoning LLM based on prompt." ellama-generate-name-by-reasoning-llm)
           (function :tag "By custom function")))
 
+(defcustom ellama-response-process-method 'streaming
+  "How to process LLM response.
+If the default streaming method is too resource-heavy, you can try other
+options."
+  :type `(choice
+	  (const :tag "Streaming" streaming)
+	  (const :tag "Async" async)
+	  (integer :tag "Skip every N messages before process one")))
+
 (defcustom ellama-define-word-prompt-template "Define %s"
   "Prompt template for `ellama-define-word'."
   :type 'string)
@@ -851,7 +861,8 @@ Defaults to md, but supports org.  Depends on `ellama-major-mode'."
   "Create new ellama session with unique id.
 Provided PROVIDER and PROMPT will be used in new session.
 If EPHEMERAL non nil new session will not be associated with any file."
-  (let* ((name (ellama-generate-name provider 'ellama prompt))
+  (let* ((dir default-directory)
+	 (name (ellama-generate-name provider 'ellama prompt))
 	 (count 1)
 	 (name-with-suffix (format "%s %d" name count))
 	 (id (if (and (not (ellama-get-session-buffer name))
@@ -872,7 +883,7 @@ If EPHEMERAL non nil new session will not be associated with any file."
 		       ellama-sessions-directory
 		       (concat id "." (ellama-get-session-file-extension)))))
 	 (session (make-ellama-session
-		   :id id :provider provider :file file-name))
+		   :id id :provider provider :file file-name :extra `(:dir ,dir)))
 	 (buffer (if file-name
 		     (progn
 		       (make-directory ellama-sessions-directory t)
@@ -881,7 +892,7 @@ If EPHEMERAL non nil new session will not be associated with any file."
     (setq ellama--current-session-id id)
     (puthash id buffer ellama--active-sessions)
     (with-current-buffer buffer
-      (setq default-directory ellama-sessions-directory)
+      (setq default-directory dir)
       (funcall ellama-major-mode)
       (setq ellama--current-session session)
       (ellama-session-mode +1))
@@ -1011,7 +1022,13 @@ If EPHEMERAL non nil new session will not be associated with any file."
 	       :provider (ellama-session-provider session)
 	       :file (ellama-session-file session)
 	       :prompt (ellama-session-prompt session)
-	       :extra extra)))
+	       :extra extra))
+	(with-current-buffer buffer
+	  (when-let* ((extra)
+		      ((plistp extra))
+		      (dir (plist-get extra :dir))
+		      ((file-exists-p dir)))
+	    (setq default-directory dir))))
       (setq ellama--current-session-id (ellama-session-id ellama--current-session))
       (puthash (ellama-session-id ellama--current-session)
 	       buffer ellama--active-sessions)
@@ -1330,7 +1347,8 @@ INSERT-REASONING is a function for reasoning insertion.
 REASONING-BUFFER is a buffer for reasoning."
   (lambda (response)
     (let ((text (plist-get response :text))
-	  (reasoning (plist-get response :reasoning)))
+	  (reasoning (plist-get response :reasoning))
+	  (tool-results (plist-get response :tool-results)))
       (funcall
        insert-text
        (concat
@@ -1338,7 +1356,7 @@ REASONING-BUFFER is a buffer for reasoning."
 	  (if
 	      (or (not ellama-output-remove-reasoning)
 		  ellama--current-session)
-	      (concat "<think>\n" reasoning)
+	      (concat "<think>\n" reasoning "\n</think>\n")
 	    (progn
 	      (with-current-buffer reasoning-buffer
 		(funcall insert-reasoning reasoning)
@@ -1348,10 +1366,94 @@ REASONING-BUFFER is a buffer for reasoning."
 		   (when ellama-reasoning-display-action-function
 		     `((ignore . (,ellama-reasoning-display-action-function)))))))
 	      nil)))
+	(when tool-results
+	  (format "\n<think>\n%s\n</think>\n"
+		  tool-results))
 	(when text
-	  (if (and reasoning ellama--current-session)
-	      (concat "</think>\n" (string-trim text))
-	    (string-trim text))))))))
+	  (string-trim text)))))))
+
+(defun ellama--error-handler (buffer errcb)
+  "Error handler function.
+BUFFER is the current ellama buffer.
+ERRCB is an error callback."
+  (lambda (_ msg)
+    (with-current-buffer buffer
+      (cancel-change-group ellama--change-group)
+      (when ellama-spinner-enabled
+	(spinner-stop))
+      (funcall errcb msg)
+      (setq ellama--current-request nil)
+      (ellama-request-mode -1))))
+
+(defun ellama--response-handler (result-handler reasoning-buffer buffer donecb errcb provider llm-prompt async filter)
+  "Response handler function.
+RESULT-HANDLER handles text insertion.
+REASONING-BUFFER used for reasoning output.
+BUFFER is the current ellama buffer.
+DONECB is a done callback.
+PROVIDER is an llm provider.
+ERRCB is an error callback.
+LLM-PROMPT is current llm prompt.
+ASYNC flag is for asyncronous requests.
+FILTER is a function that's applied to (partial) response strings before they're
+inserted into the BUFFER."
+  (lambda (response)
+    (let ((text (plist-get response :text))
+	  (reasoning (plist-get response :reasoning))
+	  (tool-result (plist-get response :tool-results)))
+      (funcall result-handler response)
+      (when (or ellama--current-session
+		(not reasoning))
+	(when (not tool-result) (kill-buffer reasoning-buffer)))
+      (if tool-result
+	  (let* ((insert-text
+		  (ellama--insert buffer (with-current-buffer buffer (if ellama--current-session
+									 (point-max)
+								       (point)))
+				  filter))
+		 (insert-reasoning
+		  (ellama--insert reasoning-buffer nil #'ellama--translate-markdown-to-org-filter))
+		 (handler (ellama--handle-partial insert-text insert-reasoning reasoning-buffer))
+		 (cnt 0)
+		 (skip-handler
+		  (lambda (request)
+		    (if (= cnt ellama-response-process-method)
+			(progn
+			  (funcall handler request)
+			  (setq cnt 0))
+		      (cl-incf cnt)))))
+	    (with-current-buffer buffer
+	      (if async
+		  (llm-chat-async
+		   provider
+		   llm-prompt
+		   (ellama--response-handler
+		    handler reasoning-buffer buffer donecb errcb provider llm-prompt async filter)
+		   (ellama--error-handler buffer errcb)
+		   t)
+		(llm-chat-streaming
+		 provider
+		 llm-prompt
+		 (if (integerp ellama-response-process-method)
+		     skip-handler handler)
+		 (ellama--response-handler
+		  handler
+		  reasoning-buffer buffer donecb errcb provider llm-prompt async filter)
+		 (ellama--error-handler buffer errcb)
+		 t))))
+	(with-current-buffer buffer
+	  (accept-change-group ellama--change-group)
+	  (when ellama-spinner-enabled
+	    (spinner-stop))
+	  (if (and (listp donecb)
+		   (functionp (car donecb)))
+	      (mapc (lambda (fn) (funcall fn text))
+		    donecb)
+	    (funcall donecb text))
+	  (when ellama-session-hide-org-quotes
+	    (ellama-collapse-org-quotes))
+	  (setq ellama--current-request nil)
+	  (ellama-request-mode -1))))))
 
 (defun ellama-stream (prompt &rest args)
   "Query ellama for PROMPT.
@@ -1416,14 +1518,18 @@ failure (with BUFFER current).
 			       (llm-chat-prompt-append-response
 				(ellama-session-prompt session)
 				prompt-with-ctx)
+			       (setf (llm-chat-prompt-tools (ellama-session-prompt session))
+				     ellama-tools-enabled)
 			       (when system
 				 (llm-chat-prompt-append-response
 				  (ellama-session-prompt session)
 				  system 'system))
 			       (ellama-session-prompt session))
 			   (setf (ellama-session-prompt session)
-				 (llm-make-chat-prompt prompt-with-ctx :context system)))
-		       (llm-make-chat-prompt prompt-with-ctx :context system))))
+				 (llm-make-chat-prompt prompt-with-ctx :context system
+						       :tools ellama-tools-enabled)))
+		       (llm-make-chat-prompt prompt-with-ctx :context system
+					     :tools ellama-tools-enabled))))
     (with-current-buffer reasoning-buffer
       (org-mode))
     (with-current-buffer buffer
@@ -1438,39 +1544,36 @@ failure (with BUFFER current).
 	  (require 'spinner)
 	  (spinner-start ellama-spinner-type))
 	(let* ((handler (ellama--handle-partial insert-text insert-reasoning reasoning-buffer))
-	       (request (llm-chat-streaming
-			 provider
-			 llm-prompt
-			 handler
-			 (lambda (response)
-			   (let ((text (plist-get response :text))
-				 (reasoning (plist-get response :reasoning)))
-			     (funcall handler response)
-			     (when (or ellama--current-session
-				       (not reasoning))
-			       (kill-buffer reasoning-buffer))
-			     (with-current-buffer buffer
-			       (accept-change-group ellama--change-group)
-			       (when ellama-spinner-enabled
-				 (spinner-stop))
-			       (if (and (listp donecb)
-					(functionp (car donecb)))
-				   (mapc (lambda (fn) (funcall fn text))
-					 donecb)
-				 (funcall donecb text))
-			       (when ellama-session-hide-org-quotes
-				 (ellama-collapse-org-quotes))
-			       (setq ellama--current-request nil)
-			       (ellama-request-mode -1))))
-			 (lambda (_ msg)
-			   (with-current-buffer buffer
-			     (cancel-change-group ellama--change-group)
-			     (when ellama-spinner-enabled
-			       (spinner-stop))
-			     (funcall errcb msg)
-			     (setq ellama--current-request nil)
-			     (ellama-request-mode -1)))
-			 t)))
+	       (request (pcase ellama-response-process-method
+			  ('async (llm-chat-async
+				   provider
+				   llm-prompt
+				   (ellama--response-handler handler reasoning-buffer buffer donecb errcb provider llm-prompt t filter)
+				   (ellama--error-handler buffer errcb)
+				   t))
+			  ('streaming (llm-chat-streaming
+				       provider
+				       llm-prompt
+				       handler
+				       (ellama--response-handler handler reasoning-buffer buffer donecb errcb provider llm-prompt nil filter)
+				       (ellama--error-handler buffer errcb)
+				       t))
+			  ((pred integerp)
+			   (let* ((cnt 0)
+				  (skip-handler
+				   (lambda (request)
+				     (if (= cnt ellama-response-process-method)
+					 (progn
+					   (funcall handler request)
+					   (setq cnt 0))
+				       (cl-incf cnt)))))
+			     (llm-chat-streaming
+			      provider
+			      llm-prompt
+			      skip-handler
+			      (ellama--response-handler handler reasoning-buffer buffer donecb errcb provider llm-prompt t filter)
+			      (ellama--error-handler buffer errcb)
+			      t))))))
 	  (with-current-buffer buffer
 	    (setq ellama--current-request request)))))))
 
